#include <cstddef>
using std::size_t;

#include <cassert>

#include <vector>
using std::vector;

#include <limits>
using std::numeric_limits;

#include <iostream>
using std::ostream;

#include <iomanip>
using std::setw;

#include <cmath>
using std::fabs;

const double NOT_ASSIGNED = numeric_limits<double>::infinity();

class grid
{
private:
	size_t Lx, Ly, N;
	vector<double> data;
	vector<double> data_tmp;
	vector<bool> fixed;
	vector<bool> dead;
	double max_error;
	
	void init_tmp(void)
	{
		assert(data_tmp.size() == 0);
		data_tmp.resize(N);
		for (size_t i = 0; i < N; ++i)
			data_tmp[i] = NOT_ASSIGNED;
	}
	
	size_t index(size_t x, size_t y) const 
	{ 
		assert(x < Lx);
		assert(y < Ly);
		return x+y*Lx; 
	}

	void relax(size_t x, size_t y, vector<double> &V)
	{
		assert(x < Lx);
		assert(y < Ly);
		const size_t i = index(x,y);
		if (!fixed[i] and !dead[i])
		{	
			double r = 0.0;
			double l = 0.0;
			double u = 0.0;
			double d = 0.0;
			unsigned short int nn = 0;
			if (x < Lx-1)
			{
				const size_t ri = index(x+1,y);
				if (!dead[ri])
				{
					r = data[ri];
					++nn;
				}
			}
			if (x > 0)
			{
				const size_t li = index(x-1,y);
				if (!dead[li])
				{
					l = data[li];
					++nn;
				}
			}
			if (y < Ly-1)
			{
				const size_t ui = index(x,y+1);
				if (!dead[ui])
				{
					u = data[ui];
					++nn;
				}
			}
			if (y > 0)
			{
				const size_t di = index(x,y-1);
				if (!dead[di])
				{
					d = data[di];
					++nn;
				}
			}
		
			assert(nn != 0);
			V[i] = (r+l+u+d)/nn;
		}
	}
	
public:
	grid(size_t Lx_, size_t Ly_) : Lx(Lx_), Ly(Ly_), N(Lx*Ly), 
	                               data(N), data_tmp(0), 
	                               fixed(N), dead(N), max_error(0) 
	{
		for (size_t i = 0; i < N; ++i)
		{
			fixed[i] = dead[i] = false;
			data[i] = 0.0;
		}
	}
	
	double& operator()(size_t x, size_t y) { return data[index(x,y)]; }
	const double& operator()(size_t x, size_t y) const { return data[index(x,y)]; }
	
	double error(void) const { return max_error; }
	void zero_error(void) { max_error = 0; }
	
	void fix(size_t x, size_t y) { fixed[index(x,y)] = true; }
	bool is_fixed(size_t x, size_t y) const { return fixed[index(x,y)]; }

	void mark_as_dead(size_t x, size_t y) { dead[index(x,y)] = true; }
	bool is_dead(size_t x, size_t y) const { return dead[index(x,y)]; }
	
	void relax(size_t x, size_t y)
	{
		const double old_value = grid::operator()(x,y);
		relax(x,y,data);
		const double err = fabs(grid::operator()(x,y) - old_value);
		if (err > max_error) max_error = err;
	}

	void relax_tentative(size_t x, size_t y)
	{
		if (data_tmp.size() == 0)
			init_tmp();
		relax(x,y,data_tmp);
	}
	
	void relax_commit(void)
	{
		for (size_t i = 0; i < N; ++i)
			if (data_tmp[i] != NOT_ASSIGNED)
			{
				const double err = fabs(data[i] - data_tmp[i]);
				if (err > max_error) max_error = err;
				data[i] = data_tmp[i];
				data_tmp[i] = NOT_ASSIGNED;
			}
	}
	
	friend ostream& operator<<(ostream &os, const grid &G)
	{
		os.precision(12);
		// grid format
		for (size_t y = 0; y < G.Ly; ++y)
		{
			for (size_t x = 0; x < G.Lx; ++x)
				os << setw(24) << G(x,y);
			os << endl;
		}	
		/* x y G(x,y) format
		for (size_t x = 0; x < G.Lx; ++x)
		{
			for (size_t y = 0; y < G.Ly; ++y)
				if (!G.is_dead(x,y))
					os << setw(12) << x
					   << setw(12) << y
					   << setw(24) << G(x,y) << endl;
		}					
		*/
		return os;
	}
};
