#include <cstddef>
using std::size_t;

#include <cassert>

#include <cstdlib>
using std::exit;

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

#include <iomanip>
using std::setw;
using std::setfill;

#include <fstream>
using std::ifstream;
using std::ofstream;

#include <string>
using std::string;

#include <sstream>
using std::stringstream;

#include "grid.hpp"

ifstream fin;
ofstream fout1;
ofstream fout2;

enum methodFlag { JACOBI, GAUSS };

int main(int argc, char* argv[])
{
	if (argc != 3)
	{
		cerr << "Usage: relax jacobi|gauss filename" << endl; 
		exit(1);
	}
	assert(argc == 3);
	const string option = argv[1];
	const string filename = argv[2];
	fin.open(filename.c_str());
	if (!fin.is_open())
	{
		cerr << "Could not open file `" << filename << "`." << endl;
		exit(1);
	}
	
	methodFlag method;
	if (option == "jacobi")
		method = JACOBI;
	else if (option == "gauss")
		method = GAUSS;
	else
	{
		cerr << "Did not recognize option `" << option << "`" << endl;
		exit(1);
	}
	
	size_t Lx, Ly;
	fin >> Lx >> Ly;
	assert(fin.good());
	grid G(Lx,Ly);
	
	// # dead cell
	// . free cell
	// 0 fixed to value zero
	// X fixed to specified value
	// + positive charge
	// - negative charge
	vector<size_t> Xcolumn;
	for (size_t y = 0; y < Ly; ++y)
	{
		Xcolumn.resize(0);
		for (size_t x = 0; x < Lx; ++x)
		{
			char c;
			fin >> c;
			assert(fin.good());
			if (c == '.') ; // do nothing
			else if (c == '#')
				G.mark_as_dead(x,y);
			else if (c == '0')
			{
				G(x,y) = 0.0;
				G.fix(x,y);
			}
			else if (c == 'X')
			{
				Xcolumn.push_back(x);
				G.fix(x,y);
			}
			else
				assert(false);
		}
		for (size_t n = 0; n < Xcolumn.size(); ++n)
			fin >> G(Xcolumn[n],y);
		assert(y == Ly-1 or fin.good());
	}

	fout1.open("output0000.dat");
	fout1 << G << endl;
	fout1.close();
	
	stringstream s;
	s << "splot[" << 0 << ":" << Lx-1 << "][" << 0 << ":" << Ly-1 << "] '";
	const string splot = s.str();
	const string matrix = "' matrix\npause 0.05";
		
	fout2.open("movie.gp");
	fout2 << "set pm3d map" << endl
	      << "set size ratio 1" << endl
	      << splot << "output0000.dat" << matrix << endl;

	unsigned int passes = 0;
	do
	{
		G.zero_error();
		if (method == JACOBI)
		{
			for (size_t x = 0; x < Lx; ++x)
				for (size_t y = 0; y < Ly; ++y)
					G.relax_tentative(x,y);
			G.relax_commit();
		}
		else if (method == GAUSS)
			for (size_t x = 0; x < Lx; ++x)
				for (size_t y = 0; y < Ly; ++y)
					G.relax(x,y);
		else
			assert(false);
		
		++passes;

		stringstream s;
		s << "output" << setw(4) << setfill('0') <<  passes << ".dat";
		fout1.open(s.str().c_str());
		fout1 << G << endl;
		fout1.close();
		
		fout2 << splot << s.str() << matrix << endl;

		cout << "Error : " << G.error() << endl;
	} while (G.error() > 1E-5 and passes < 10000);

	if (passes == 10000)
		cout << "Did not converge after 10000 passes" << endl;
	else
		cout << "Converged in " << passes << " passes" << endl;

	fout2.close();	
	return 0;
}
